from pytest import raises

from numpy import arange, isin, bool_

from hypothesis import given, assume, example
from hypothesis.strategies import integers, one_of

from ..slice import Slice
from ..integer import Integer
from ..ellipsis import ellipsis
from ..ndindex import asshape
from .helpers import check_same, slices, prod, shapes, iterslice, assert_equal

def test_slice_args():
    # Test the behavior when not all three arguments are given
    # TODO: Incorporate this into the normal slice tests
    raises(TypeError, lambda: slice())
    raises(TypeError, lambda: Slice())
    raises(TypeError, lambda: Slice(1.0))
    # See docstring of operator_index()
    raises(TypeError, lambda: Slice(True))
    raises(TypeError, lambda: Slice(bool_(True)))

    S = Slice(1)
    assert S == Slice(S) == Slice(None, 1) == Slice(None, 1, None) == Slice(None, 1, None)
    assert S.raw == slice(None, 1, None)
    assert S.args == (S.start, S.stop, S.step)

    S = Slice(0, 1)
    assert S == Slice(S) == Slice(0, 1, None)
    assert S.raw == slice(0, 1, None)
    assert S.args == (S.start, S.stop, S.step)

    S = Slice(0, 1, 2)
    assert S == Slice(S)
    assert S.raw == slice(0, 1, 2)
    assert S.args == (S.start, S.stop, S.step)

def test_slice_exhaustive():
    for n in range(100):
        a = arange(n)
        for start, stop, step in iterslice(one_two_args=False):
            check_same(a, slice(start, stop, step))

@given(slices(), integers(0, 100))
def test_slice_hypothesis(s, size):
    a = arange(size)
    check_same(a, s)

def test_slice_len_exhaustive():
    for args in iterslice():
        try:
            S = Slice(*args)
        except ValueError:
            continue
        try:
            l = len(S)
        except ValueError:
            # No maximum
            l = 10000

        m = -1
        for n in range(20):
            a = arange(n)
            L = len(a[S.raw])
            assert L <= l, S
            m = max(L, m)
        if l != 10000:
            assert m == l, S
        else:
            # If there is no maximum, the size of the slice should increase
            # with larger arrays.
            assert len(arange(30)[S.raw]) > m, S

        # TODO
        # if l == 0:
        #     # There should only be one canonical length 0 slice
        #     assert s == Slice(0, 0)

@given(slices())
def test_slice_len_hypothesis(s):
    try:
        S = Slice(s)
    except ValueError: # pragma: no cover
        assume(False)
    try:
        l = len(S)
    except ValueError:
        # No maximum
        l = 10000

    m = -1
    for n in range(20):
        a = arange(n)
        L = len(a[S.raw])
        assert L <= l, (S, n)
        m = max(L, m)
    if l != 10000:
        assert m == l, S
    else:
        # If there is no maximum, the size of the slice should increase
        # with larger arrays.
        assert len(arange(30)[S.raw]) > m, S

def test_slice_args_reduce_no_shape():
    S = Slice(1).reduce()
    assert S == Slice(None, 1).reduce() == Slice(0, 1, None).reduce() == Slice(0, 1).reduce() == Slice(0, 1, 1)

    S = Slice(0, 1).reduce()
    assert S == Slice(0, 1, None).reduce() == Slice(0, 1, 1)

def test_slice_reduce_no_shape_exhaustive():
    slices = {}
    A = [arange(n) for n in range(30)]
    for args in iterslice():
        try:
            S = Slice(*args)
        except ValueError:
            continue

        # Check the conditions stated by the Slice.reduce() docstring
        reduced = S.reduce()
        assert reduced.start != None
        if S.start != None and S.start >= 0:
            assert reduced.start >= 0
        assert reduced.step != None
        if S.step is not None:
            assert abs(reduced.step) <= abs(S.step)
        if reduced.stop is None:
            assert S.stop is None
        assert reduced.reduce() == reduced, S

        B = []
        for a in A:
            check_same(a, S.raw, ndindex_func=lambda a, x: a[x.reduce().raw])
            B.append(tuple(a[reduced.raw]))
        B = tuple(B)
        # Test that Slice.reduce gives a canonical result, i.e., if any two
        # slices always give the same sub-arrays, they reduce to the same thing
        if B in slices:
            assert slices[B] == reduced, f"{S} reduced to {reduced}, but should be equal to {slices[B]}"
        else:
            slices[B] = reduced


@given(slices(), one_of(integers(0, 100), shapes))
def test_slice_reduce_no_shape_hypothesis(s, shape):
    if isinstance(shape, int):
        a = arange(shape)
    else:
        a = arange(prod(shape)).reshape(shape)
    try:
        S = Slice(s)
    except ValueError: # pragma: no cover
        assume(False)

    # The axis argument is tested implicitly in the Tuple.reduce test. It is
    # difficult to test here because we would have to pass in a Tuple to
    # check_same.
    check_same(a, S.raw, ndindex_func=lambda a, x: a[x.reduce().raw])

    # Check the conditions stated by the Slice.reduce() docstring
    reduced = S.reduce()
    assert reduced.start != None
    if S.start != None and S.start >= 0:
        assert reduced.start >= 0
    assert reduced.step != None
    if S.step is not None:
        assert abs(reduced.step) <= abs(S.step)
    if reduced.stop is None:
        assert S.stop is None
    assert reduced.reduce() == reduced, S

def test_slice_reduce_exhaustive():
    for n in range(30):
        slices = {}
        a = arange(n)
        for args in iterslice():
            try:
                S = Slice(*args)
            except ValueError:
                continue

            reduced = S.reduce((n,))

            # Check the conditions stated by the Slice.reduce() docstring
            # TODO: Factor this out so we can also test it in the tuple reduce
            # tests.

            # len() should not raise after calling reduce() with a shape
            L = len(reduced)
            # len() should be exact after calling reduce() with a shape
            assert L == len(a[reduced.raw]), (S, n)
            assert reduced.start >= 0
            # We cannot require stop > 0 because if stop = None and step < 0, the
            # only equivalent stop that includes 0 is negative.
            assert reduced.stop != None
            if S.step != None and S.step < 0:
                if reduced.stop < 0:
                    assert reduced.stop == -n - 1
                    assert 0 in a[reduced.raw], (S, n)
                    assert L > 1
            else:
                assert reduced.stop >= 0, (S, n)
            assert reduced.step != None
            if S.step != None:
                assert abs(reduced.step) <= abs(S.step)
            if S.stop != None and S.stop >= 0:
                # Test that stop is as close to start as possible for the
                # given step (the "as possible" is checked by uniqueness
                # below).
                if L not in [0, 1]:
                    if reduced.step > 0:
                        assert reduced.stop <= S.stop, (S, n)
                    else:
                        assert reduced.stop >= S.stop
            if L == 1:
                assert reduced == Slice(reduced.start, reduced.start+1, 1)

            check_same(a, S.raw, ndindex_func=lambda a, x: a[x.reduce((n,)).raw])
            B = tuple(a[reduced.raw])
            # Test that Slice.reduce gives a canonical result, i.e., if any two
# slices always give the same sub-arrays, they reduce to the same thing
            if B in slices:
                assert slices[B] == reduced, f"{S} reduced to {reduced}, but should be equal to {slices[B]} for shape {n}"
            else:
                slices[B] = reduced

@example(slice(None, None, -1), 2)
@example(slice(-10, 11, 3), 10)
@example(slice(-1, 3, -3), 10)
@given(slices(), one_of(integers(0, 100), shapes))
def test_slice_reduce_hypothesis(s, shape):
    if isinstance(shape, int):
        a = arange(shape)
    else:
        a = arange(prod(shape)).reshape(shape)

    try:
        S = Slice(s)
    except ValueError: # pragma: no cover
        assume(False)

    # The axis argument is tested implicitly in the Tuple.reduce test. It is
    # difficult to test here because we would have to pass in a Tuple to
    # check_same.
    check_same(a, S.raw, ndindex_func=lambda a, x: a[x.reduce(shape).raw])

    # Check the conditions stated by the Slice.reduce() docstring
    try:
        reduced = S.reduce(shape)
    except IndexError:
        # shape == ()
        return

    shape = asshape(shape)
    n = shape[0]

    # len() should not raise after calling reduce() with a shape
    L = len(reduced)
    # len() should be exact after calling reduce() with a shape
    assert L == len(a[reduced.raw]), (S, n)
    assert reduced.start >= 0
    # We cannot require stop > 0 because if stop = None and step < 0, the
    # only equivalent stop that includes 0 is negative.
    assert reduced.stop != None
    if S.step != None and S.step < 0:
        if reduced.stop < 0:
            assert reduced.stop == -n - 1
            if a.size != 0:
                assert a[0] in a[reduced.raw], (S, n)
            assert L > 1
    else:
        assert reduced.stop >= 0, (S, n)
    assert reduced.step != None
    if S.step != None:
        assert abs(reduced.step) <= abs(S.step)
    if S.stop != None and S.stop >= 0:
        # Test that stop is as close to start as possible for the
        # given step (the "as possible" is checked by uniqueness
        # in the exhaustive test).
        if L not in [0, 1]:
            if reduced.step > 0:
                assert reduced.stop <= S.stop, (S, n)
            else:
                assert reduced.stop >= S.stop
    if L == 1:
        assert reduced == Slice(reduced.start, reduced.start+1, 1)

def test_slice_newshape_exhaustive():
    def raw_func(a, idx):
        return a[idx].shape

    def ndindex_func(a, index):
        return index.newshape(shape)

    def assert_equal(raw_shape, newshape):
        assert raw_shape == newshape

    for n in range(10):
        shape = n
        a = arange(n)

        for sargs in iterslice():
            try:
                S = Slice(*sargs)
            except ValueError:
                continue

            check_same(a, S.raw, raw_func=raw_func, ndindex_func=ndindex_func,
                   assert_equal=assert_equal)

def test_slice_as_subindex_slice_exhaustive():
    # We have to restrict the range of the exhaustive test to get something
    # that finishes in a reasonable amount of time (~30 seconds, vs. 30
    # minutes for the original ranges).

    # a = arange(10)
    # for sargs in iterslice():
    #     for indexargs in iterslice():

    a = arange(5)
    for sargs in iterslice((-5, 5), (-5, 5), (-5, 5), one_two_args=False):
        for indexargs in iterslice((-5, 5), (-5, 5), (-5, 5), one_two_args=False):

            try:
                S = Slice(*sargs)
            except ValueError:
                continue

            try:
                Index = Slice(*indexargs)
            except ValueError:
                continue

            try:
                Subindex = S.as_subindex(Index)
            except NotImplementedError:
                continue

            aS = a[S.raw]
            aindex = a[Index.raw]
            asubindex = aindex[Subindex.raw]

            assert_equal(asubindex, aS[isin(aS, aindex)])

            subindex2 = Index.as_subindex(S)
            asubindex2 = aS[subindex2.raw]
            assert_equal(asubindex2, asubindex)

def test_slice_as_subindex_integer_exhaustive():
    a = arange(10)
    for sargs in iterslice():
        for i in range(-10, 10):

            try:
                S = Slice(*sargs)
            except ValueError:
                continue

            Index = Integer(i)

            empty = False
            try:
                Subindex = S.as_subindex(Index)
            except NotImplementedError:
                continue
            except ValueError as e:
                assert "do not intersect" in e.args[0]
                empty = True

            aS = a[S.raw]
            aindex = a[i]

            if empty:
                assert not isin(aS, aindex).any()
                assert not isin(aindex, aS).any()
                with raises(ValueError, match="do not intersect"):
                    Index.as_subindex(S)
            else:
                asubindex = aindex[Subindex.raw]

                assert_equal(asubindex.flatten(), aS[isin(aS, aindex)])

                subindex2 = Index.as_subindex(S)
                asubindex2 = aS[subindex2.raw]
                assert_equal(asubindex2, asubindex)

def test_slice_as_subindex_ellipsis_exhaustive():
    a = arange(10)
    for sargs in iterslice():
        try:
            S = Slice(*sargs)
        except ValueError:
            continue

        Index = ellipsis()

        try:
            Subindex = S.as_subindex(Index)
        except NotImplementedError:
            continue

        aS = a[S.raw]
        aindex = a[...]

        asubindex = aindex[Subindex.raw]

        assert_equal(asubindex.flatten(), aS[isin(aS, aindex)])

        try:
            subindex2 = Index.as_subindex(S)
        except NotImplementedError:
            continue
        asubindex2 = aS[subindex2.raw]
        assert_equal(asubindex2, asubindex)

def test_slice_isempty_exhaustive():
    for args in iterslice():
        try:
            S = Slice(*args)
        except ValueError:
            continue

        isempty = S.isempty()

        aempty = True
        for n in range(30):
            a = arange(n)

            aS = a[S.raw]
            if aS.size != 0:
                if isempty:
                    raise AssertionError(f"Slice s = {S}.isempty() gave True, a[s] is not empty for a = range({n}).")
                else:
                    aempty = False
            # isempty() should always give the correct result for a specific
            # array shape
            assert S.isempty(n) == (aS.size == 0)

        assert isempty == aempty, S

@example(slice(None, None, None), ())
@given(slices(), one_of(shapes, integers(0, 10)))
def test_slice_isempty_hypothesis(s, shape):
    if isinstance(shape, int):
        a = arange(shape)
    else:
        a = arange(prod(shape)).reshape(shape)

    try:
        S = Slice(s)
    except (IndexError, ValueError): # pragma: no cover
        assume(False)

    def raw_func(a, s):
        return a[s].size == 0

    def ndindex_func(a, S):
        return S.isempty(), S.isempty(shape)

    def assert_equal(raw_empty, ndindex_empty):
        isempty, isempty_shape = ndindex_empty

        # If isempty is True then a[t] should be empty
        if isempty:
            assert raw_empty, (S, shape)
        # We cannot test the converse with hypothesis. isempty may be False
        # but a[s] could still be empty for this specific a (e.g., if a is
        # already itself empty).

        # If isempty is true with no shape it should be true for a specific
        # shape. The converse is not true because the indexed array could be
        # empty.
        if isempty:
            assert isempty_shape, (S, shape)

        # isempty() should always give the correct result for a specific
        # array after reduction
        assert isempty_shape == raw_empty, (S, shape)

    check_same(a, s, raw_func=raw_func, ndindex_func=ndindex_func,
               assert_equal=assert_equal, same_exception=False)



import setuptools
import versioneer

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ndindex",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Quansight Labs",
    description="A Python library for manipulating indices of ndarrays.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://quansight-labs.github.io/ndindex/",
    packages=['ndindex', 'ndindex.tests'],
    license="MIT",
    install_requires=[
        "numpy",
        "sympy",
    ],
    tests_require=[
        'pytest',
        'hypothesis',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)



"""
Custom script to run the doctests

This runs the doctests but ignores trailing ``` in Markdown documents.

Running this separately from pytest also allows us to not include the doctests
in the coverage. It also allows us to force a separate namespace for each
docstring's doctest, which the pytest doctest integration does not allow.

TODO: Make these tests also run with pytest, but still keeping them out of the
coverage.

"""

import sys
import unittest
import glob
import os
from contextlib import contextmanager
from doctest import DocTestRunner, DocFileSuite, DocTestSuite, NORMALIZE_WHITESPACE

@contextmanager
def patch_doctest():
    """
    Context manager to patch the doctester

    The doctester must be patched
    """
    orig_run = DocTestRunner.run

    def run(self, test, **kwargs):
        for example in test.examples:
            # Remove ```
            example.want = example.want.replace('```\n', '')
            example.exc_msg = example.exc_msg and example.exc_msg.replace('```\n', '')

        return orig_run(self, test, **kwargs)

    try:
        DocTestRunner.run = run
        yield
    finally:
        DocTestRunner.run = orig_run

DOCS = os.path.realpath(os.path.join(__file__, os.path.pardir, os.path.pardir,
                                     os.pardir, 'docs'))
MARKDOWN = glob.glob(os.path.join(DOCS, '**', '*.md'), recursive=True)
RST = glob.glob(os.path.join(DOCS, '**', '*.rst'), recursive=True)
README = os.path.realpath(os.path.join(__file__, os.path.pardir, os.path.pardir,
                                     os.pardir, 'README.md'))
def load_tests(loader, tests, ignore):
    for mod in sys.modules:
        if mod.startswith('ndindex'):
            # globs={} makes the doctests not include module names
            tests.addTests(DocTestSuite(sys.modules[mod], globs={},
                                        optionflags=NORMALIZE_WHITESPACE))
    tests.addTests(DocFileSuite(*MARKDOWN, *RST, README,
                                optionflags=NORMALIZE_WHITESPACE,
                                module_relative=False))
    return tests

def doctest():
    with patch_doctest():
        return unittest.main(module='ndindex.tests.doctest', exit=False).result

if __name__ == '__main__':
    # TODO: Allow specifying which doctests to run at the command line
    doctest()



__all__ = []

from .ndindex import ndindex

__all__ += ['ndindex']

from .slice import Slice

__all__ += ['Slice']

from .integer import Integer

__all__ += ['Integer']

from .tuple import Tuple

__all__ += ['Tuple']

from .ellipsis import ellipsis

__all__ += ['ellipsis']

from .newaxis import Newaxis

__all__ += ['Newaxis']

from .integerarray import IntegerArray

__all__ += ['IntegerArray']

from .booleanarray import BooleanArray

__all__ += ['BooleanArray']

from .chunking import ChunkSize

__all__ += ['ChunkSize']

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



from itertools import product

from numpy import arange, array, intp, empty

from hypothesis import given, example
from hypothesis.strategies import integers, one_of

from pytest import raises

from ..ndindex import ndindex
from ..tuple import Tuple
from ..integer import Integer
from .helpers import check_same, Tuples, prod, short_shapes, iterslice

def test_tuple_constructor():
    # Test things in the Tuple constructor that are not tested by the other
    # tests below.
    raises(ValueError, lambda: Tuple((1, 2, 3)))
    raises(ValueError, lambda: Tuple(0, (1, 2, 3)))

    # Test NotImplementedError behavior for Tuples with arrays split up by
    # slices, ellipses, and newaxes.
    raises(NotImplementedError, lambda: Tuple(0, slice(None), [0]))
    raises(NotImplementedError, lambda: Tuple([0], slice(None), [0]))
    raises(NotImplementedError, lambda: Tuple([0], slice(None), [0]))
    raises(NotImplementedError, lambda: Tuple(0, ..., [0]))
    raises(NotImplementedError, lambda: Tuple([0], ..., [0]))
    raises(NotImplementedError, lambda: Tuple([0], ..., [0]))
    raises(NotImplementedError, lambda: Tuple(0, None, [0]))
    raises(NotImplementedError, lambda: Tuple([0], None, [0]))
    raises(NotImplementedError, lambda: Tuple([0], None, [0]))
    # Make sure this doesn't raise
    Tuple(0, slice(None), 0)
    Tuple(0, ..., 0)
    Tuple(0, None, 0)

def test_tuple_exhaustive():
    # Exhaustive tests here have to be very limited because of combinatorial
    # explosion.
    a = arange(2*2*2).reshape((2, 2, 2))
    types = {
        slice: lambda: iterslice((-1, 1), (-1, 1), (-1, 1), one_two_args=False),
        # slice: _iterslice,
        int: lambda: ((i,) for i in range(-3, 3)),
        type(...): lambda: ()
    }

    for t1, t2, t3 in product(types, repeat=3):
        for t1_args in types[t1]():
            for t2_args in types[t2]():
                for t3_args in types[t3]():
                    idx1 = t1(*t1_args)
                    idx2 = t2(*t2_args)
                    idx3 = t3(*t3_args)

                    idx = (idx1, idx2, idx3)
                    # Disable the same exception check because there could be
                    # multiple invalid indices in the tuple, and for instance
                    # numpy may give an IndexError but we would give a
                    # TypeError because we check the type first.
                    check_same(a, idx, same_exception=False)
                    try:
                        index = Tuple(*idx)
                    except (IndexError, ValueError):
                        pass
                    else:
                        assert index.has_ellipsis == (type(...) in (t1, t2, t3))

@given(Tuples, short_shapes)
def test_tuples_hypothesis(t, shape):
    a = arange(prod(shape)).reshape(shape)
    check_same(a, t, same_exception=False)

@given(Tuples, short_shapes)
def test_ellipsis_index(t, shape):
    a = arange(prod(shape)).reshape(shape)
    # Don't know if there is a better way to test ellipsis_idx
    def ndindex_func(a, index):
        return a[ndindex((*index.raw[:index.ellipsis_index], ...,
                          *index.raw[index.ellipsis_index+1:])).raw]

    check_same(a, t, ndindex_func=ndindex_func)

@example((True, 0, False), 1)
@example((..., None), ())
@given(Tuples, one_of(short_shapes, integers(0, 10)))
def test_tuple_reduce_no_shape_hypothesis(t, shape):
    if isinstance(shape, int):
        a = arange(shape)
    else:
        a = arange(prod(shape)).reshape(shape)

    index = Tuple(*t)

    check_same(a, index.raw, ndindex_func=lambda a, x: a[x.reduce().raw],
               same_exception=False)

    reduced = index.reduce()
    if isinstance(reduced, Tuple):
        assert len(reduced.args) != 1
        assert reduced == () or reduced.args[-1] != ...

@example((..., None), ())
@example((..., empty((0, 0), dtype=bool)), (0, 0))
@example((empty((0, 0), dtype=bool), 0), (0, 0, 1))
@example((array([], dtype=intp), 0), (0, 0))
@example((array([], dtype=intp), array(0)), (0, 0))
@example((array([], dtype=intp), [0]), (0, 0))
@example((0, 1, ..., 2, 3), (2, 3, 4, 5, 6, 7))
@example((0, slice(None), ..., slice(None), 3), (2, 3, 4, 5, 6, 7))
@example((0, ..., slice(None)), (2, 3, 4, 5, 6, 7))
@example((slice(None, None, -1),), (2,))
@example((..., slice(None, None, -1),), (2, 3, 4))
@given(Tuples, one_of(short_shapes, integers(0, 10)))
def test_tuple_reduce_hypothesis(t, shape):
    if isinstance(shape, int):
        a = arange(shape)
    else:
        a = arange(prod(shape)).reshape(shape)

    index = Tuple(*t)

    check_same(a, index.raw, ndindex_func=lambda a, x: a[x.reduce(shape).raw],
               same_exception=False)

    try:
        reduced = index.reduce(shape)
    except IndexError:
        pass
    else:
        if isinstance(reduced, Tuple):
            assert len(reduced.args) != 1
            assert reduced == () or reduced.args[-1] != ...
        # TODO: Check the other properties from the Tuple.reduce docstring.

def test_tuple_reduce_explicit():
    # Some aspects of Tuple.reduce are hard to test as properties, so include
    # some explicit tests here.

    # (Before Index, shape): After index
    tests = {
        # Make sure redundant slices are removed
        (Tuple(0, ..., slice(0, 3)), (5, 3)): Integer(0),
        (Tuple(slice(0, 5), ..., 0), (5, 3)): Tuple(..., Integer(0)),
        # Ellipsis is removed if unnecessary
        (Tuple(0, ...), (2, 3)): Integer(0),
        (Tuple(0, 1, ...), (2, 3)): Tuple(Integer(0), Integer(1)),
        (Tuple(..., 0, 1), (2, 3)): Tuple(Integer(0), Integer(1)),
    }

    for (before, shape), after in tests.items():
        reduced = before.reduce(shape)
        assert reduced == after, (before, shape)

        a = arange(prod(shape)).reshape(shape)
        check_same(a, before.raw, ndindex_func=lambda a, x:
                   a[x.reduce(shape).raw])

@example((slice(0, 0),), 2)
@example((0, slice(0, 0)), (1, 2))
@given(Tuples, one_of(short_shapes, integers(0, 10)))
def test_tuple_isempty_hypothesis(t, shape):
    if isinstance(shape, int):
        a = arange(shape)
    else:
        a = arange(prod(shape)).reshape(shape)

    T = Tuple(*t)

    try:
        ndindex(t).isempty(shape)
    except NotImplementedError:
        return
    except IndexError:
        pass

    def raw_func(a, t):
        return a[t].size == 0

    def ndindex_func(a, T):
        return T.isempty(), T.isempty(shape)

    def assert_equal(raw_empty, ndindex_empty):
        isempty, isempty_shape = ndindex_empty

        # If isempty is True then a[t] should be empty
        if isempty:
            assert raw_empty, (T, shape)
        # We cannot test the converse with hypothesis. isempty may be False
        # but a[t] could still be empty for this specific a (e.g., if a is
        # already itself empty).

        # If isempty is true with no shape it should be true for a specific
        # shape. The converse is not true because the indexed array could be
        # empty.
        if isempty:
            assert isempty_shape, (T, shape)

        # isempty() should always give the correct result for a specific
        # array after reduction
        assert isempty_shape == raw_empty, (T, shape)

    check_same(a, t, raw_func=raw_func, ndindex_func=ndindex_func,
               assert_equal=assert_equal, same_exception=False)



from .ndindex import NDIndex, asshape

class Newaxis(NDIndex):
    """
    Represents a `np.newaxis` (i.e., `None`) index.

    `Newaxis` adds a shape 1 dimension to the array. If a `Newaxis` is inside
    of a tuple index, it adds a shape 1 dimension at that location in the
    index.

    For example, if `a` has shape `(2, 3)`, then `a[newaxis]` has shape `(1,
    2, 3)`, `a[:, newaxis]` has shape `(2, 1, 3)`, and so on.

    >>> from ndindex import Newaxis
    >>> from numpy import arange
    >>> a = arange(0,6).reshape(2,3)
    >>> a[Newaxis().raw].shape
    (1, 2, 3)
    >>> a[:, Newaxis().raw, :].shape
    (2, 1, 3)

    Using `Newaxis().raw` as an index is equivalent to using `numpy.newaxis`.

    .. note::

       Unlike the NumPy `newaxis`, `Newaxis` is the type, not the object (the
       name is uppercase to avoid conflicting with the NumPy type). Use
       `Newaxis()`, `ndindex(np.newaxis)`, or `ndindex(None)` to create the
       object. In most ndindex contexts, `np.newaxis` or `None` can be used
       instead of `Newaxis()`, for instance, when creating a `Tuple` object.
       Also unlike `None`, `Newaxis()` is not singletonized, so you should not
       use `is` to compare it. See the document on :any:`type-confusion` for
       more details.

    """
    def _typecheck(self):
        return ()

    @property
    def raw(self):
        return None

    def reduce(self, shape=None, axis=0):
        """
        Reduce a `Newaxis` index

        There is no other index that is equivalent to a newaxis index by
        itself, so `Newaxis().reduce()` always returns `Newaxis()` unchanged.

        >>> from ndindex import Newaxis
        >>> Newaxis().reduce()
        Newaxis()

        See Also
        ========

        .NDIndex.reduce
        .Tuple.reduce
        .Slice.reduce
        .Integer.reduce
        .ellipsis.reduce
        .IntegerArray.reduce
        .BooleanArray.reduce

        """
        if shape is not None:
            shape = asshape(shape)
        return self

    def newshape(self, shape):
        # The docstring for this method is on the NDIndex base class
        shape = asshape(shape)

        # reduce will raise IndexError if it should be raised
        self.reduce(shape)

        return (1,) + shape

    def isempty(self, shape=None):
        if shape is not None:
            return 0 in self.newshape(shape)

        return False

    def __eq__(self, other):
        return other is None or isinstance(other, Newaxis)

    def __hash__(self):
        return super().__hash__()



from numpy import arange, int64, isin, bool_

from pytest import raises

from hypothesis import given, example
from hypothesis.strategies import integers, one_of

from ..integer import Integer
from ..slice import Slice
from .helpers import check_same, ints, prod, shapes, iterslice, assert_equal

def test_integer_args():
    zero = Integer(0)
    assert zero.raw == 0
    idx = Integer(int64(0))
    assert idx == zero
    assert idx.raw == 0
    assert isinstance(idx.raw, int)
    assert Integer(zero) == zero

    raises(TypeError, lambda: Integer(1.0))
    # See the docstring of operator_index()
    raises(TypeError, lambda: Integer(True))
    raises(TypeError, lambda: Integer(bool_(True)))

def test_integer_exhaustive():
    a = arange(10)
    for i in range(-12, 12):
        check_same(a, i)


@given(ints(), integers(5, 100))
def test_integer_hypothesis(i, size):
    a = arange(size)
    check_same(a, i)


def test_integer_len_exhaustive():
    for i in range(-12, 12):
        idx = Integer(i)
        assert len(idx) == 1


@given(ints())
def test_integer_len_hypothesis(i):
    idx = Integer(i)
    assert len(idx) == 1


def test_integer_reduce_exhaustive():
    a = arange(10)
    for i in range(-12, 12):
        check_same(a, i, ndindex_func=lambda a, x: a[x.reduce((10,)).raw])

        try:
            reduced = Integer(i).reduce(10)
        except IndexError:
            pass
        else:
            assert reduced.raw >= 0


@given(ints(), shapes)
def test_integer_reduce_hypothesis(i, shape):
    a = arange(prod(shape)).reshape(shape)
    # The axis argument is tested implicitly in the Tuple.reduce test. It is
    # difficult to test here because we would have to pass in a Tuple to
    # check_same.
    check_same(a, i, ndindex_func=lambda a, x: a[x.reduce(shape).raw])

    try:
        reduced = Integer(i).reduce(shape)
    except IndexError:
        pass
    else:
        assert reduced.raw >= 0

def test_integer_reduce_no_shape_exhaustive():
    a = arange(10)
    for i in range(-12, 12):
        check_same(a, i, ndindex_func=lambda a, x: a[x.reduce().raw])

@given(ints(), shapes)
def test_integer_reduce_no_shape_hypothesis(i, shape):
    a = arange(prod(shape)).reshape(shape)
    check_same(a, i, ndindex_func=lambda a, x: a[x.reduce().raw])

def test_integer_newshape_exhaustive():
    shape = 5
    a = arange(shape)

    def raw_func(a, idx):
        return a[idx].shape

    def ndindex_func(a, index):
        return index.newshape(shape)

    def assert_equal(raw_shape, newshape):
        assert raw_shape == newshape

    for i in range(-10, 10):
        check_same(a, i, raw_func=raw_func, ndindex_func=ndindex_func,
                   assert_equal=assert_equal)

def test_integer_as_subindex_slice_exhaustive():
    for n in range(10):
        a = arange(n)
        for i in range(-10, 10):
            try:
                a[i]
            except IndexError:
                continue

            for indexargs in iterslice():
                idx = Integer(i)

                try:
                    Index = Slice(*indexargs)
                except ValueError:
                    continue

                empty = False
                try:
                    Subindex = idx.as_subindex(Index)
                except NotImplementedError:
                    continue
                except ValueError as e:
                    assert "do not intersect" in e.args[0]
                    empty = True

                aidx = a[idx.raw]
                aindex = a[Index.raw]
                if empty:
                    assert not isin(aidx, aindex).any()
                    assert not isin(aindex, aidx).any()
                    with raises(ValueError, match="do not intersect"):
                        Index.as_subindex(idx)
                else:
                    asubindex = aindex[Subindex.raw]

                    assert_equal(asubindex.flatten(), aidx[isin(aidx, aindex)])

                    subindex2 = Index.as_subindex(idx)
                    asubindex2 = aidx[subindex2.raw]
                    assert_equal(asubindex2, asubindex)

def test_integer_isempty_exhaustive():
    for i in range(-10, 10):
        idx = Integer(i)

        isempty = idx.isempty()

        for n in range(30):
            a = arange(n)

            exception = False
            try:
                aidx = a[idx.raw]
            except IndexError:
                exception = True
            else:
                if aidx.size != 0:
                    # If a[i] doesn't give an index error, it should always be nonempty
                    assert not isempty
            # isempty() should always give the correct result for a specific
            # array shape
            try:
                isemptyn = idx.isempty(n)
            except IndexError:
                if not exception:
                    raise AssertionError(f"idx.isempty(n) raised but a[idx] did not (idx = {idx}, n = {n}).")
            else:
                if exception:
                    raise AssertionError(f"a[idx] raised but idx.isempty(n) did not (idx = {idx}, n = {n}).")
                assert isemptyn == (aidx.size == 0)

@example(1, (2, 0))
@given(ints(), one_of(shapes, integers(0, 10)))
def test_integer_isempty_hypothesis(i, shape):
    if isinstance(shape, int):
        a = arange(shape)
    else:
        a = arange(prod(shape)).reshape(shape)

    index = Integer(i)

    def raw_func(a, idx):
        return a[idx].size == 0

    def ndindex_func(a, index):
        return index.isempty(), index.isempty(shape)

    def assert_equal(raw_empty, ndindex_empty):
        isempty, isempty_shape = ndindex_empty

        # Since i is an integer, it should never be unconditionally empty
        assert not isempty
        # We cannot test the converse with hypothesis. isempty may be False
        # but a[i] could still be empty for this specific a (e.g., if a is
        # already itself empty).

        # isempty() should always give the correct result for a specific
        # array after reduction
        assert isempty_shape == raw_empty, (index, shape)

    check_same(a, i, raw_func=raw_func, ndindex_func=ndindex_func,
               assert_equal=assert_equal, same_exception=False)
